/*
 *  LittleOS
 *  Copyright (C) 1998 Lacroix Pascal (placr@mygale.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * memory.c - very simple memory management (no free function)
 */

#include <kernel.h>
#include <paging.h>
#include <string.h>
#include <bitops.h>

extern end;
extern void die();

ulong total_memory, total_pages;
ulong base_memory;

static ulong mem_map[32768];		/* 32768*8= 1Mb --> supports 4GB */
					/* one bit for each page */

#define ALIGN4(x) (x) = ( ((x) + 3) & ~3 )

static inline void __freepage(ulong page_no)
{
        clear_bit(page_no, mem_map);
}

static inline void __allocpage(ulong page_no)
{
	if (test_bit(page_no, &mem_map)) D("warning: page is already allocated");

	set_bit(page_no, mem_map);
}

static inline char __is_pagefree(ulong page_no)
{
        return test_bit(page_no, mem_map);
}

static inline ulong __find_freepage(ulong start_addr)
{
        ulong index;

        index = start_addr / PAGE_SIZE;
      	while (!__is_pagefree(index) && index<32768*32) {
		index++;
        }
        if (__is_pagefree(index)) return index*PAGE_SIZE;
        else return 0;	/* page 0 is never used */
}

/* allocate a page (4096 bytes) */
ulong alloc_page(void)
{
	ulong tmp;
        tmp = PAGE_ROUND_UP(base_memory);
        tmp += PAGE_SZ;
        base_memory = tmp;
        ALIGN4(base_memory);
        memset((void*) tmp, 0, PAGE_SZ);
        if (base_memory>total_memory) return 0;
        return tmp;
}


/* allocate a block of `size' */
/* NOTE: this function should not be used, because paging is not
  really working */
void *kmalloc(ulong size)
{
	ulong tmp;
        tmp = base_memory;
        base_memory += size;
        ALIGN4(base_memory);
        if (base_memory>total_memory) {
        	printk("Out of memory (in %s line %d)\n", __FILE__, __LINE__);
                die();
        }
        memset((void*) tmp, 0, size);
        return (void *) tmp;
}

void init_memory(void)
{
	ulong i, j;
        ulong kernel_end, tmp;
        ulong starta, enda;

	/* read total of memory from CMOS */
	outb(0x70, 0x30);
        total_memory = inb(0x71) * 1024L;
        outb(0x70, 0x31);
	total_memory += (inb(0x71) << 8) * 1024L;
        total_memory += 0x100000;	/* adjust (first Mb) */

        /* number of pages of PAGE_SIZE (=4096) bytes */
        total_pages = total_memory/PAGE_SIZE;

        /* begin of free memory after kernel */
	base_memory = (ulong) &end;
        base_memory = PAGE_ROUND_UP(base_memory);
        kernel_end = base_memory + kernel_base;

        printk("\eRMemory:\eN %d Kb total, %d Kb free (kernel takes %d Kb)\n", total_memory/1024, (total_memory-base_memory)/1024,
        	base_memory/1024);

        /* init memory map (must be tested) */
        __allocpage(0);		/* first page is never free */
        for (i = 1; i < total_pages; i++) {
        	if ((i >= 0xA0000/PAGE_SIZE && i < 0x100000) ||
                	(i >= kernel_base/PAGE_SZ && i <= kernel_end/PAGE_SIZE)) __allocpage(i);
                else
                __freepage(i);
        }

}

